"""
Testing hysop.field.continuous.Field
"""
import hysop
print hysop.__file__
from hysop.fields.continuous import Field
from hysop.domain.box import Box
from hysop.tools.parameters import Discretization
import numpy as np
from hysop.fields.tests.func_for_tests import func_scal_1, func_scal_2, \
    func_vec_1, func_vec_2, func_vec_3, func_vec_4, func_vec_5, func_vec_6
from numpy import allclose

d3D = Discretization([33, 33, 33])
d2D = Discretization([33, 33])
nbc = 4


def test_continuous():
    """ Basic continuous field construction """

    dom = Box()
    cf = Field(dom)
    assert cf.nb_components == 1
    assert cf.domain is dom
    cf = Field(dom, is_vector=True)
    assert cf.nb_components == dom.dimension
    assert cf.domain is dom


def test_analytical():
    """Test formula"""
    dom = Box()
    formula = lambda x, y, z: (x * y * z, 1.0, -x * y * z)
    caf = Field(dom, formula, vectorize_formula=True)
    assert caf.value(0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3.) == (6., 1., -6.)


def test_analytical_reset():
    """Test formula"""
    dom = Box()
    formula = lambda x, y, z: (x * y * z, 1.0, -x * y * z)
    caf = Field(dom, formula, vectorize_formula=True)
    assert caf.value(0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3.) == (6., 1., -6.)
    form2 = lambda x, y, z, dt: (x * y * z, 1.0, -x * y * z * dt)
    caf.set_formula(form2, vectorize_formula=True)
    assert caf.value(0., 0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3., 4.) == (6., 1., -24.)


def test_discretization():
    """Test multiple discretizations"""
    dom = Box()
    csf = Field(dom)
    cvf = Field(dom, is_vector=True)
    r3d = Discretization([33, 33, 17])
    topo = dom.create_topology(r3d)
    csf.discretize(topo)
    cvf.discretize(topo)
    assert np.equal(csf.discreteFields[topo].resolution,
                    csf.discreteFields[topo].data[0].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[0].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[1].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[2].shape).all()


# Non-Vectorized formula for a scalar
def test_analytical_field_1():
    box = Box()
    topo = box.create_topology(discretization=d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_scal_1)
    ref = Field(box)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = id(cafd.data[0])
    caf.initialize()
    refd.data = func_scal_1(refd.data, *(coords + (0.,)))
    assert allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_scal_1(refd.data, *(coords + (time,)))
    assert allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids


# Vectorized formula
def test_analytical_field_2():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_scal_2, vectorize_formula=True)
    ref = Field(box)
    cafd = caf.discretize(topo)
    ids = id(cafd.data[0])
    refd = ref.discretize(topo)
    caf.initialize()
    refd.data = func_scal_1(refd.data, *(coords + (0.,)))
    assert allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_scal_1(refd.data, *(coords + (time,)))
    assert allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids


# Non-Vectorized formula for a vector
def test_analytical_field_3():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_1, is_vector=True)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_1(refd.data, *(coords + (time,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Vectorized formula for a vector
def test_analytical_field_4():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_2, is_vector=True,
                vectorize_formula=True)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_1(refd.data, *(coords + (time,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a vector, with extra-arguments
def test_analytical_field_5():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_3, is_vector=True)
    theta = 0.3
    caf.set_formula_parameters(theta)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_3(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_3(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Vectorized formula for a vector, with extra-arguments
def test_analytical_field_6():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_4, is_vector=True,
                vectorize_formula=True)
    theta = 0.3
    caf.set_formula_parameters(theta)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_3(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_3(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a field with nb_components
# different from domain dim and  with extra-arguments
def test_analytical_field_7():
    box = Box()
    topo = box.create_topology(d3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_5, nb_components=nbc)
    theta = 0.3
    caf.set_formula_parameters(theta)
    ref = Field(box, nb_components=nbc)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * nbc
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])

    caf.initialize()
    refd.data = func_vec_5(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_5(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a 2D field with nb_components
# different from domain dim and  with extra-arguments
def test_analytical_field_8():
    box = Box(dimension=2, length=[1., 1.], origin=[0., 0.])
    topo = box.create_topology(d2D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_6, nb_components=nbc)
    theta = 0.3
    caf.set_formula_parameters(theta)
    ref = Field(box, nb_components=nbc)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = [0, ] * nbc
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_6(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(time=time)
    refd.data = func_vec_6(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a vector, initialization on several
# topologies.
def test_analytical_field_9():
    box = Box()
    topo = box.create_topology(d3D)
    res2 = Discretization([65, 33, 65], [1, 1, 1])
    topo2 = box.create_topology(res2, dim=2)
    coords = topo.mesh.coords
    coords2 = topo2.mesh.coords
    caf = Field(box, formula=func_vec_1, is_vector=True)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    cafd2 = caf.discretize(topo2)
    refd2 = ref.discretize(topo2)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd2.data[i])
        # init on topo2
    caf.initialize(topo=topo2)
    refd2.data = func_vec_1(refd2.data, *(coords2 + (0.,)))
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd2[i], refd2.data[i])
        assert id(cafd2.data[i]) == ids[i]
        assert not allclose(cafd[i], refd.data[i])
    caf.initialize(topo=topo)
    for i in xrange(caf.nb_components):
        assert allclose(cafd[i], refd.data[i])


# Non-Vectorized formula for a vector, initialization on several
# topologies.
def test_analytical_field_10():
    box = Box()
    topo = box.create_topology(d3D)
    res2 = Discretization([65, 33, 65], [1, 1, 1])
    topo2 = box.create_topology(res2, dim=2)
    coords = topo.mesh.coords
    coords2 = topo2.mesh.coords
    caf = Field(box, formula=func_vec_1, is_vector=True)
    ref = Field(box, is_vector=True)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    cafd2 = caf.discretize(topo2)
    refd2 = ref.discretize(topo2)
    ids = [0, ] * 3
    for i in xrange(3):
        ids[i] = id(cafd2.data[i])
    # init on all topos
    caf.initialize()
    refd2.data = func_vec_1(refd2.data, *(coords2 + (0.,)))
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nb_components):
        assert allclose(cafd2[i], refd2.data[i])
        assert allclose(cafd[i], refd.data[i])
        assert id(cafd2.data[i]) == ids[i]


# This may be useful to run mpi tests
if __name__ == "__main__":
    test_continuous()
    test_analytical()
    test_analytical_reset()
    test_discretization()
    test_analytical_field_1()
    test_analytical_field_2()
    test_analytical_field_3()
    test_analytical_field_4()
    test_analytical_field_5()
    test_analytical_field_6()
    test_analytical_field_7()
    test_analytical_field_8()
